@@ -8,6 +8,7 @@ from datetime import timedelta
 
 import backoff
 import humps
+import ijson
 import requests
 import singer
 import singer.metrics as metrics
@@ -216,6 +217,21 @@ class Stream():
     def __init__(self, config=None):
         self.config = config
 
+    def send_request_get_results(self, req):
+        resp = session.send(req)
+
+        if 'Too Many Requests' in resp.reason:
+            retry_after = 30
+            LOGGER.info("Rate limit reached. Sleeping for %s seconds",
+                        retry_after)
+            time.sleep(retry_after)
+            raise Server42xRateLimitError(resp.reason)
+
+        resp.raise_for_status()
+
+        dec = humps.decamelize(resp.json())
+        return dec
+
     @backoff.on_exception(backoff.expo, (requests.exceptions.RequestException, Server42xRateLimitError),
                           max_tries=5,
                           giveup=lambda e: e.response is not None and 400 <= e.
@@ -251,19 +267,7 @@ class Stream():
         LOGGER.info("%s %s %s", request_kwargs['method'],
                     request_kwargs['url'], request_kwargs['params'])
 
-        resp = session.send(req)
-
-        if 'Too Many Requests' in resp.reason:
-            retry_after = 30
-            LOGGER.info("Rate limit reached. Sleeping for %s seconds",
-                        retry_after)
-            time.sleep(retry_after)
-            raise Server42xRateLimitError(resp.reason)
-
-        resp.raise_for_status()
-
-        dec = humps.decamelize(resp.json())
-        return dec
+        return self.send_request_get_results(req)
 
     def get_bookmark(self, state, stream, default, key=None):
         if (state is None) or ('bookmarks' not in state):
@@ -351,10 +355,11 @@ class Stream():
                                        'inclusion', 'available')
 
         # For period stream adjust schema for time period
-        if hasattr(self, 'period') and self.period == 'hourRange':
-            mdata.pop(('properties', 'day'))
-        elif hasattr(self, 'period') and self.period == 'dayRange':
-            mdata.pop(('properties', 'hour'))
+        if self.replication_key == 'day' or self.replication_key == 'hour':
+            if hasattr(self, 'period') and self.period == 'hourRange':
+                mdata.pop(('properties', 'day'))
+            elif hasattr(self, 'period') and self.period == 'dayRange':
+                mdata.pop(('properties', 'hour'))
 
         return metadata.to_list(mdata)
 
@@ -364,6 +369,8 @@ class Stream():
     def transform(self, record):
         return humps.decamelize(record)
 
+    LOG_PROGRESS_PERCENTAGE_INTERVAL = 5
+
     def sync_substream(self, state, parent, sub_stream, parent_response):
         bookmark_date = self.get_bookmark(state, sub_stream.name,
                                           self.config.get('start_date'),
@@ -388,7 +395,9 @@ class Stream():
                     parent_response = parent_response[i:len(parent_response)]
                     continue
 
-        for record in parent_response:
+
+        next_log_progress_percentage = 0
+        for index, record in enumerate(parent_response):
             try:
                 with metrics.record_counter(
                         sub_stream.name) as counter, Transformer(
@@ -434,6 +443,12 @@ class Stream():
             # All events for all parents processed; can removed last processed
             self.update_bookmark(state=state, stream=sub_stream.name, bookmark_value=record.get(parent.key_properties[0]), bookmark_key="last_processed")
             self.update_bookmark(state=state, stream=sub_stream.name, bookmark_value=strftime(new_bookmark), bookmark_key=sub_stream.replication_key)
+
+            progress_percentage = float(index) / len(parent_response) * 100
+            if progress_percentage > next_log_progress_percentage:
+                LOGGER.info("Finished syncing %s percentage of sub_stream for parent %s's sub_stream %s data", progress_percentage, parent.name, sub_stream.name)
+                next_log_progress_percentage += self.LOG_PROGRESS_PERCENTAGE_INTERVAL
+
         # After processing for all parent ids we can remove our resumption state
         state.get('bookmarks').get(sub_stream.name).pop('last_processed')
         update_currently_syncing(state, None)
@@ -459,6 +474,35 @@ class Stream():
             raise TypeError("lookback_window '{}' is not numeric. Check your configuration".format(lookback_window))
         return int(lookback_window)
 
+class LazyAggregationStream(Stream):
+    def send_request_get_results(self, req):
+        with session.send(req, stream=True) as resp:
+            if 'Too Many Requests' in resp.reason:
+                retry_after = 30
+                LOGGER.info("Rate limit reached. Sleeping for %s seconds",
+                            retry_after)
+                time.sleep(retry_after)
+                raise Server42xRateLimitError(resp.reason)
+
+            resp.raise_for_status()
+
+            for item in ijson.items(resp.raw, 'results.item'):
+                yield humps.decamelize(item)
+
+    def sync(self, state, start_date=None, key_id=None):
+        stream_response = self.request(self.name, json=self.get_body()) or []
+
+        if STREAMS.get(SUB_STREAMS.get(self.name)):
+            sub_stream = STREAMS.get(SUB_STREAMS.get(self.name))(self.config)
+        else:
+            sub_stream = None
+
+        if stream_response and sub_stream and sub_stream.is_selected():
+            self.sync_substream(state, self, sub_stream, stream_response)
+
+        update_currently_syncing(state, None)
+        return (self.stream, stream_response)
+
 class EventsBase(Stream):
     DATE_WINDOW_SIZE = 1
     key_properties = ['visitor_id', 'account_id', 'server', 'remote_ip']
@@ -576,7 +620,7 @@ class FeatureEvents(EventsBase):
         }
 
 
-class Events(Stream):
+class Events(LazyAggregationStream):
     name = "events"
     DATE_WINDOW_SIZE = 1
     key_properties = ['visitor_id', 'account_id', 'server', 'remote_ip']
@@ -602,7 +646,7 @@ class Events(Stream):
 
         period = self.config.get('period')
         body = self.get_body(period, ts)
-        events = self.request(self.name, json=body).get('results') or []
+        events = self.request(self.name, json=body) or []
         update_currently_syncing(state, None)
         return self.stream, events
 
@@ -639,9 +683,10 @@ class PollEvents(Stream):
         super().__init__(config=config)
         self.config = config
         self.period = config.get('period')
-        self.replication_key = "day" if self.period == 'dayRange' else "hour"
+        self.replication_key = 'browser_time'
 
     def get_body(self, period, first):
+        sort = humps.camelize(self.replication_key)
         return {
             "response": {
                 "mimeType": "application/json"
@@ -657,7 +702,7 @@ class PollEvents(Stream):
                         }
                     }
                 }, {
-                    "sort": [self.replication_key]
+                    "sort": [sort]
                 }]
             }
         }
@@ -713,26 +758,36 @@ class GuideEvents(EventsBase):
     name = "guide_events"
     key_properties = ['visitor_id', 'account_id', 'server_name', 'remote_ip']
 
+    def __init__(self, config):
+        super().__init__(config=config)
+        self.config = config
+        self.period = config.get('period')
+        self.replication_key = 'browser_time'
+
     def get_body(self, key_id, period, first):
+        sort = humps.camelize(self.replication_key)
         return {
             "response": {
                 "mimeType": "application/json"
             },
             "request": {
-                "pipeline": [{
-                    "source": {
-                        "guideEvents": {
-                            "guideId": key_id
-                        },
-                        "timeSeries": {
-                            "period": period,
-                            "first": first,
-                            "last": "now()"
+                "pipeline": [
+                    {
+                        "source": {
+                            "guideEvents": {
+                                "guideId": key_id
+                            },
+                            "timeSeries": {
+                                "period": period,
+                                "first": first,
+                                "last": "now()"
+                                }
                         }
+                    },
+                    {
+                        "sort": [sort]
                     }
-                }, {
-                    "sort": [self.replication_key]
-                }]
+                ]
             }
         }
 
@@ -902,7 +957,7 @@ class VisitorHistory(Stream):
         return super().transform(record)
 
 
-class Visitors(Stream):
+class Visitors(LazyAggregationStream):
     name = "visitors"
     replication_method = "INCREMENTAL"
     replication_key = "lastupdated"
